/* to get sys_errlist declared */
#define _BSD_SOURCE

#include <stdlib.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_ether.h>
#include <net/if_arp.h>
#include <netpacket/packet.h>
#include <unistd.h>
#include <time.h>
#include <sys/ioctl.h>
#include <resolv.h>
#include <errno.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <ctype.h>
#include "uint16.h"
#include "fmt.h"
#include "ip4.h"
#include "ip6.h"
#include "case.h"
#include "buffer.h"

#ifndef __dietlibc__
#define __attribute__(x)
#endif


static void die(const char* message) {
  buffer_puts(buffer_2,message);
  buffer_puts(buffer_2,": ");
  buffer_puts(buffer_2,strerror(errno));
  buffer_putnlflush(buffer_2);
  exit(1);
}

unsigned long long arp;
unsigned long long http,ftp,mysql,nfs,nntp,https,nntps,donkey,torrent,ssh;

struct ftpconnection {
  char srcip[16];
  char dstip[16];
  uint32 window;
  uint16 srcport,dstport;
  char dsrcip[16],ddstip[16];
  uint16 srcdataport,dstdataport;
  struct ftpconnection* next;
  char buf[100];	/* for the FTP PORT/PASV/EPRT/EPSV command */
  int n,expecting;
}* root;

static void ftppkt(const char* srcip,const char* dstip,uint16 srcport,uint16 dstport,
		   const unsigned char* tcp,const unsigned char* max) {
  struct ftpconnection** r;

  {
    char a[IP6_FMT];
    char b[IP6_FMT];
    a[fmt_ip6c(a,srcip)]=0;
    b[fmt_ip6c(b,dstip)]=0;
//    printf("FTP: %s %d  ->  %s %d\n",a,srcport,b,dstport);
  }

//  if (dstport != 21) return;
  r=&root;
  while (*r) {
    if (byte_equal((*r)->srcip,16,srcip) &&
	byte_equal((*r)->dstip,16,dstip) &&
	(*r)->srcport == srcport &&
	(*r)->dstport == dstport)
      break;
    r=&((*r)->next);
  }
//  printf("found %p; flags %x\n",(*r),tcp[13]);
  if ((tcp[13]&2)==2 || !*r) {		/* syn or not there */
//    printf("ftp syn\n");
    if (*r) {
      puts("abort: syn but context already there");
      return;
    }
    *r=malloc(sizeof(**r));
    if (!*r) {
      puts("abort: malloc failed");
      return;
    }
    byte_zero(*r,sizeof(struct ftpconnection));

    byte_copy((*r)->srcip,16,srcip);
    byte_copy((*r)->dstip,16,dstip);
    (*r)->srcport=srcport;
    (*r)->dstport=dstport;
    (*r)->window=uint32_read(tcp+4);
  }
  if (!*r) {
//    puts("abort: no context found");
    return;
  }
//  printf("using %p\n",*r);
  {
    int i;
    const unsigned char* payload=tcp+(tcp[12]>>4)*4;
    if (payload>max) return;
//    printf("got tcp: %lu (%d bytes: \"%.*s\")\n",uint32_read(tcp+4),max-payload,max-payload,payload);
    for (i=0; payload+i<max; ++i) {
      if ((*r)->n<99)
	(*r)->buf[(*r)->n++]=payload[i];
      if (payload[i]=='\r' || payload[i]=='\n') {
	(*r)->buf[(*r)->n]=0;
	if ((*r)->n) {
//	  printf("got ftp %s: %s\n",dstport==21?"request":"response",(*r)->buf);
	  if (case_starts((*r)->buf,"PORT ") && (*r)->n>=16) {
	    /* PORT 160,45,44,199,128,29 */
	    struct ftpconnection* x=(*r);
	    unsigned long tmp[6];
	    int j,kaputt;
	    char* y=x->buf+5;
	    kaputt=0;
	    for (j=0; j<6; ++j) {
	      int k=scan_ulong(y,tmp+j);
	      if (y[k]!=(j==5?0:',') || tmp[j]>255) { kaputt=1; break; }
	      y+=k+1;
	    }
	    if (!kaputt) {
	      byte_copy(x->dsrcip,16,x->dstip);
	      byte_copy(x->ddstip,12,V4mappedprefix);
	      x->ddstip[12]=tmp[0];
	      x->ddstip[13]=tmp[1];
	      x->ddstip[14]=tmp[2];
	      x->ddstip[15]=tmp[3];
	      x->dstdataport=(tmp[4]<<8)+tmp[5];
	      x->srcdataport=0;
	      x->expecting=1;

#if 0
	      {
		char a[IP6_FMT];
		char b[IP6_FMT];
		a[fmt_ip6c(a,x->dsrcip)]=0;
		b[fmt_ip6c(b,x->ddstip)]=0;
		printf("expecting ftp connection: %s %d  -->  %s %d\n",a,x->srcdataport,b,x->dstdataport);
	      }
#endif
	    }
	  } else if (case_starts((*r)->buf,"EPRT ") && (*r)->n>=18) {
	    /* EPRT |1|160.45.44.199|32809| */
	    /* EPRT |2|::1|1234| */
	    struct ftpconnection* x=(*r);
	    char* y=x->buf+5;
	    char sep=*y;
	    int kaputt=0;
	    if (y[1]=='1' || y[1]=='2') {
	      int tmp;
	      unsigned long port;
	      y+=3;
	      if (y[tmp=scan_ip6(y,x->ddstip)]!=sep) kaputt=1;
	      if (!kaputt) {
		y+=tmp+1;
		if (y[tmp=scan_ulong(y,&port)]!=sep) kaputt=1;
	      }
	      if (!port || port>0xffff) kaputt=1;
	      if (!kaputt) {
		y+=tmp+1;
		x->dstdataport=port;
		byte_copy(x->dsrcip,16,x->dstip);
		x->srcdataport=0;
		x->expecting=1;

#if 0
		{
		  char a[IP6_FMT];
		  char b[IP6_FMT];
		  a[fmt_ip6c(a,x->dsrcip)]=0;
		  b[fmt_ip6c(b,x->ddstip)]=0;
		  printf("expecting ftp connection: %s %d  -->  %s %d\n",a,x->srcdataport,b,x->dstdataport);
		}
#endif
	      }
	    }
	  } else if (case_starts((*r)->buf,"229 ")) {
	    /* 229 Passive Mode OK (|||37079|) */
	    struct ftpconnection* x=(*r);
	    char* y=x->buf+4;
	    unsigned long tmp,port;
	    int kaputt=0;
	    while (*y && !isdigit(*y)) ++y;
	    if (y[tmp=scan_ulong(y,&port)]!=y[-1]) kaputt=1;
	    if (!port || port>0xffff) kaputt=1;
	    if (!kaputt) {
	      byte_copy(x->dsrcip,16,x->dstip);
	      byte_copy(x->ddstip,16,x->srcip);
	      x->dstdataport=port;
	      x->srcdataport=0;
	      x->expecting=1;
#if 0
	      {
		char a[IP6_FMT];
		char b[IP6_FMT];
		a[fmt_ip6c(a,x->dsrcip)]=0;
		b[fmt_ip6c(b,x->ddstip)]=0;
		printf("expecting ftp connection: %s %d  -->  %s %d\n",a,x->srcdataport,b,x->dstdataport);
	      }
#endif
	    }
	  } else if (case_starts((*r)->buf,"227 ")) {
	    /* 227 Passive Mode OK (160,45,44,196,144,223) */
	    struct ftpconnection* x=(*r);
	    char* y=x->buf+5;
	    unsigned long tmp[6];
	    int j,kaputt;
	    while (*y && !isdigit(*y)) ++y;
	    kaputt=0;
	    for (j=0; j<6; ++j) {
	      int k=scan_ulong(y,tmp+j);
	      if ((j<5 && y[k]!=',') || tmp[j]>255) { kaputt=1; break; }
	      y+=k+1;
	    }
	    if (!kaputt) {
	      byte_copy(x->dsrcip,16,x->dstip);
	      byte_copy(x->ddstip,12,V4mappedprefix);
	      x->ddstip[12]=tmp[0];
	      x->ddstip[13]=tmp[1];
	      x->ddstip[14]=tmp[2];
	      x->ddstip[15]=tmp[3];
	      x->dstdataport=(tmp[4]<<8)+tmp[5];
	      x->srcdataport=0;
	      x->expecting=1;

#if 0
	      {
		char a[IP6_FMT];
		char b[IP6_FMT];
		a[fmt_ip6c(a,x->dsrcip)]=0;
		b[fmt_ip6c(b,x->ddstip)]=0;
		printf("expecting ftp connection: %s %d  -->  %s %d\n",a,x->srcdataport,b,x->dstdataport);
	      }
#endif
	    }
	  }
	}
	(*r)->n=0;
	if (payload[i]=='\r') ++i;
      }
    }
  }
}

static void ip(int v,const unsigned char* buf,unsigned long len) {	/* rfc791 */
  unsigned char srcip[16];
  unsigned char dstip[16];
  int sport,dport;
  const char* next;
  int typeofs;
  if (v==4) {
    byte_copy(srcip,12,V4mappedprefix);
    byte_copy(srcip+12,4,buf+12);
    byte_copy(dstip,12,V4mappedprefix);
    byte_copy(dstip+12,4,buf+16);
    next=(const char*)buf+(buf[0]&0xf)*4;
    if ((buf[0]&0xf0) != 0x40) return;
    typeofs=9;
  } else if (v==6) {
    byte_copy(srcip,16,buf+8);
    byte_copy(dstip,16,buf+8+16);
    if ((buf[0]&0xf0) != 0x60) return;
    next=(const char*)buf+8+2*16;
    typeofs=6;
  } else
    exit(1);

  sport=uint16_read_big(next);
  dport=uint16_read_big(next+2);

  if (buf[typeofs]==6 || buf[typeofs]==17) {
    if (sport == 80 || dport == 80)
      http+=len;
    else if (sport == 443 || dport == 443)
      https+=len;
    else if (sport == 119 || dport == 119)
      nntp+=len;
    else if (sport == 563 || dport == 563)
      nntps+=len;
    if (sport == 22 || dport == 22)
      ssh+=len;
    else if (sport == 3306 || dport == 3306)
      mysql+=len;
    else if (sport == 21 || dport == 21 || sport == 20 || dport == 20) {
      ftppkt(srcip,dstip,sport,dport,next,buf+len);
      ftp+=len;
    } else if (sport == 2049 || dport == 2049)
      nfs+=len;
    else if ((sport >= 4662 && sport <= 4669) || (dport >= 4662 && dport <= 4669))
      donkey+=len;
    else if ((sport >= 6881 && sport <= 6889) || (dport >= 6881 && dport <= 6889))
      torrent+=len;
    else {
      struct ftpconnection* f=root;

#if 0
      {
	char a[IP6_FMT];
	char b[IP6_FMT];
	a[fmt_ip6c(a,srcip)]=0;
	b[fmt_ip6c(b,dstip)]=0;
	printf("got tcp: %s %d  -->  %s %d\n",a,sport,b,dport);
      }
#endif

      while (f) {
	if (f->expecting) {
	  if (byte_equal(f->dsrcip,16,srcip) &&
	      byte_equal(f->ddstip,16,dstip) &&
	      f->dstdataport==dport &&
	      (f->srcdataport==sport || f->srcdataport==0)) {
	    if (next[13]&5) f->expecting=0;
	    ftp+=len;
	  } else if (byte_equal(f->dsrcip,16,dstip) &&
	      byte_equal(f->ddstip,16,srcip) &&
	      f->dstdataport==sport &&
	      (f->srcdataport==dport || f->srcdataport==0)) {
	    if (next[13]&5) f->expecting=0;
	    ftp+=len;
	  }
	}
	f=f->next;
      }
    }
  }
}

static void layer2(const unsigned char* const buf,const unsigned int proto,unsigned long len) {
  switch (proto) {
  case 0x8864: /* pppoe */ break;
  case 0x0800: ip(4,buf,len); break;
  case 0x86DD: ip(6,buf,len); break;
  case 0x0806: arp+=len; break;
  case 0: break;
  }
}

int main(int argc,char* argv[]) {
  int sock,len;
  char buf[16*1024];
  struct sockaddr_ll sl,from;
  socklen_t fromlen=sizeof from;
  const char* interface="eth0";
  struct ifreq ifr;
  time_t now;
  unsigned long long in,out;
  char myname[1024];
  struct winsize ws;
  int row;

  row=1;
  if (ioctl(1,TIOCGWINSZ,&ws)==-1)
    ws.ws_row=25;

  gethostname(myname,sizeof myname);

  in=out=0;

  for (;;) {
    int i;
    if ((i=getopt(argc,argv,"i:evn"))==-1) break;
    switch (i) {
    case 0:
      buffer_puts(buffer_2,"non-option ");
      buffer_puts(buffer_2,argv[optind]);
      buffer_putnlflush(buffer_2);
      break;
    case 'i': interface=optarg; break;
    default: buffer_putsflush(buffer_2,"usage: tcpdump [-nev] [-i ifname]"); return 0; break;
    }
  }
  buffer_puts(buffer_2,"using interface ");
  buffer_puts(buffer_2,interface);
  buffer_putnlflush(buffer_2);
  sock=socket(PF_PACKET,SOCK_RAW,ETH_P_ALL);
  if (sock==-1) die("socket");
  sl.sll_family=AF_PACKET;
  sl.sll_protocol=htons(ETH_P_ALL);
  sl.sll_ifindex=if_nametoindex(interface);
  if (bind(sock,(struct sockaddr*)&sl,sizeof sl)==-1) die("bind");
  {
    strncpy(ifr.ifr_name,interface,IFNAMSIZ);
    if (ioctl(sock,SIOCGIFHWADDR, &ifr)==-1) die("SIOCGIFINDEX");
    switch (ifr.ifr_hwaddr.sa_family) {
    case ARPHRD_ETHER:
    case ARPHRD_LOOPBACK:
    case ARPHRD_IEEE80211: // WLAN
    case ARPHRD_PPP:
      break;
    default:
      buffer_puts(buffer_2,"unknown layer 2: ");
      buffer_putulong(buffer_2,ifr.ifr_hwaddr.sa_family);
      buffer_putnlflush(buffer_2);
      exit(1);
    }
  }
  time(&now);
  while ((len=recvfrom(sock,buf,sizeof buf,0,(struct sockaddr*)&from,&fromlen))>0) {
    struct timeval tv;
    struct tm *t;

    if (ifr.ifr_hwaddr.sa_family==ARPHRD_LOOPBACK && from.sll_pkttype!=PACKET_OUTGOING)
      /* Linux gives every packet twice on loopback.  Only look at
       * outgoing packets */
      continue;
    if (from.sll_pkttype==PACKET_OUTGOING)
      out+=len;
    else
      in+=len;
    ioctl(sock,SIOCGSTAMP,&tv);
    t=localtime(&tv.tv_sec);

    switch (ifr.ifr_hwaddr.sa_family) {
      case ARPHRD_ETHER:
      case ARPHRD_LOOPBACK:
	layer2(buf+14,uint16_read_big(buf+12),len-14);
	break;
    }
    if (tv.tv_sec != now) {
      if (tv.tv_sec<now && tv.tv_sec>now-10) continue;
      now=tv.tv_sec;
      if (!--row) {
	row=ws.ws_row-3;
	printf("%-9s KBin KBout   nfs  http https  nntp nntps mysql   ftp   ssh  ed2k    bt\n"
	       "--------------------------------------------------------------------------------\n",myname);
      }
      printf("%02d:%02d:%02d %5llu %5llu %5llu %5llu %5llu %5llu %5llu %5llu %5llu %5llu %5llu %5llu\n",t->tm_hour,t->tm_min,t->tm_sec,in/1024,out/1024,nfs/1024,http/1024,https/1024,nntp/1024,nntps/1024,mysql/1024,ftp/1024,ssh/1024,donkey/1024,torrent/1024);
      in=out=0;
      nfs=http=https=ftp=mysql=arp=nntp=nntps=donkey=torrent=ssh=0;
    }
  }
  die("recv");
  return 0;
}
